import io
import re

def __export_remote_functions():

    def substitute_vars(string, vars: dict):
        # Define a regular expression for placeholders
        placeholder_pattern = re.compile(r"\$\{?(\w+)(?:-(\w+))?\}?")
        # Define a function to replace the placeholders with keyword arguments
        def replace_placeholder(match):
            # Get the variable name and the default value from the match object
            var_name = match.group(1)
            default_value = match.group(2)
            # Check if the variable name is in the keyword arguments
            if var_name in vars:
                # Return the value of the keyword argument
                return vars[var_name]
            else:
                # Check if there is a default value
                if default_value is not None:
                    # Return the default value
                    return default_value
                else:
                    # Raise an exception if there is no default value
                    raise ValueError(f"Missing keyword argument: {var_name}")
        # Return the string with the placeholders replaced
        return placeholder_pattern.sub(replace_placeholder, string)

    def process_cp2k_macro(fp):
        # Define a regular expression for @SET directive with case insensitive flag
        set_pattern = re.compile(r"@SET\s+(\w+)\s+(.+)", re.IGNORECASE) # Added re.IGNORECASE flag
        # Initialize an empty dictionary for variables and a list for output lines
        variables = {}
        output_lines = []
        # Read line by line from the input file object
        for line in fp:
            # Strip whitespace and comments
            line = line.strip()
            line = line.split("#")[0]
            # Skip empty lines
            if not line:
                continue
            # Match @SET directive
            set_match = set_pattern.match(line)
            if set_match:
                # Get the variable name and value
                var_name = set_match.group(1)
                var_value = set_match.group(2).strip()

                # Check if the variable name is valid
                if var_name[0].isdigit():
                    # Raise an exception if the variable name starts with a number
                    raise ValueError(f"Invalid variable name: {var_name}")
                else:
                    # Assign the value to the variable in the dictionary
                    variables[var_name] = var_value
                # Skip the @SET line
                continue

            # Append the line to the output list
            output_lines.append(line)

        # Return the variables dictionary and the output list as a single string with newline characters
        return variables, "\n".join(output_lines)

    def parse_cp2k_input(fp):
        # Initialize an empty dictionary and a stack
        output = {}
        stack = []
        # Open the input file and read line by line
        for line in fp:
            # Strip whitespace and comments
            line = line.strip()
            line = line.split("#")[0]
            # Skip empty lines
            if not line:
                continue
            # Check if the line starts with &
            if line.startswith("&"):
                # Get the keyword name and strip the trailing whitespace
                keyword = line[1:].strip()
                # Split the keyword by whitespace and check if the first token is END
                tokens = keyword.split()
                if tokens[0] and tokens[0].upper() == "END":
                    # Pop the last section from the stack
                    stack.pop()
                else:
                    # If the stack is empty, add the keyword to the output dictionary
                    if not stack:
                        output[keyword] = {}
                        stack.append(keyword)
                    else:
                        # If the stack is not empty, get the current section from the output dictionary
                        current_section = output
                        for section in stack:
                            current_section = current_section[section]
                        # Add the keyword to the current section as a sub-dictionary
                        current_section[keyword] = {}
                        stack.append(keyword)
                # Continue to the next line
                continue
            # Split the line by whitespace
            tokens = line.split()
            # Get the value name and value
            value_name = tokens[0]
            value = " ".join(tokens[1:])
            # Get the current section from the output dictionary
            current_section = output
            for section in stack:
                current_section = current_section[section]
            # Add the value name and value to the current section
            current_section[value_name] = value
        return output

    # TODO: handle coords
    def load_cp2k_input(fp):
        variables, processed_text = process_cp2k_macro(fp)
        substituted_text = substitute_vars(processed_text, variables)
        return parse_cp2k_input(io.StringIO(substituted_text))

    def loads_cp2k_input(text):
        return load_cp2k_input(io.StringIO(text))

    # TODO: handle coords
    def dumps_cp2k_input(input_dict):
        # Initialize an empty list for output lines
        output_lines = []
        # Define a helper function to recursively dump the sections and values
        def dump_section(section_dict, indent=0):
            # Loop through the keys and values in the section dictionary
            for key, value in section_dict.items():
                # Check if the value is a sub-dictionary
                if isinstance(value, dict):
                    # Add a line with the section name and indentation
                    output_lines.append(" " * indent + f"&{key}")
                    # Recursively dump the sub-section with increased indentation
                    dump_section(value, indent + 3)
                    # Add a line with the end of section and indentation
                    output_lines.append(" " * indent + "&END")
                else:
                    # Add a line with the value name and value and indentation
                    output_lines.append(" " * indent + f"{key}  {value}")
        # Dump the input dictionary using the helper function
        dump_section(input_dict)
        # Return the output list as a single string with newline characters
        return "\n".join(output_lines)

    def dump_cp2k_input(input_dict, fp):
        fp.write(dumps_cp2k_input(input_dict))

    return dump_cp2k_input, dumps_cp2k_input, load_cp2k_input, loads_cp2k_input,


(
    dump_cp2k_input, dumps_cp2k_input,
    load_cp2k_input, loads_cp2k_input,
) = __export_remote_functions()



from ai2_kit.core.artifact import ArtifactDict

from typing import List, Tuple, Optional
from ase import Atoms

import ase.io
import os


def __export_remote_functions():
    """workaround for cloudpickle issue"""

    class DataFormat:
        # customize data format
        CP2K_OUTPUT_DIR = 'cp2k/output_dir'
        VASP_OUTPUT_DIR = 'vasp/output_dir'
        LAMMPS_OUTPUT_DIR = 'lammps/output_dir'
        DEEPMD_OUTPUT_DIR = 'deepmd/output_dir'
        DEEPMD_MODEL = 'deepmd/model'
        DEEPMD_NPY = 'deepmd/npy'
        LASP_LAMMPS_OUT_DIR ='lasp+lammps/output_dir'

        # data format of dpdata
        CP2K_OUTPUT = 'cp2k/output'
        VASP_XML = 'vasp/xml'

        # data format of ase
        EXTXYZ = 'extxyz'
        VASP_POSCAR = 'vasp/poscar'


    def get_data_format(artifact: dict) -> Optional[str]:
        """
        Get (or guess) data type from artifact dict
        Note: The reason of using dict instead of Artifact is Artifact is not pickleable
        """
        url = artifact.get('url')
        assert isinstance(url, str), f'url must be str, got {type(url)}'

        file_name = os.path.basename(url)
        format = artifact.get('format')
        if format and isinstance(format, str):
            return format  # TODO: validate format
        if file_name.endswith('.xyz'):
            return DataFormat.EXTXYZ
        if 'POSCAR' in file_name:
            return DataFormat.VASP_POSCAR
        return None


    def artifacts_to_ase_atoms(artifacts: List[ArtifactDict], type_map: List[str]) -> List[Tuple[ArtifactDict, Atoms]]:
        results = []
        for a in artifacts:
            data_format = get_data_format(a)  # type: ignore
            url = a['url']
            if data_format == DataFormat.VASP_POSCAR:
                atoms_list = ase.io.read(url, ':', format='vasp')
            elif data_format == DataFormat.EXTXYZ:
                atoms_list = ase.io.read(url, ':', format='extxyz')
            else:
                raise ValueError(f'unsupported data format: {data_format}')
            results.extend((a, atoms) for atoms in atoms_list)
        return results


    def ase_atoms_to_cp2k_input_data(atoms: Atoms) -> Tuple[List[str], List[List[float]]]:
        coords = [atom.symbol + ' ' + ' '.join(str(x) for x in atom.position) for atom in atoms] # type: ignore
        cell = [list(row) for row in atoms.cell]  # type: ignore
        return (coords, cell)


    def convert_to_lammps_input_data(systems: List[ArtifactDict], base_dir: str, type_map: List[str]):
        data_files = []
        atoms_list = artifacts_to_ase_atoms(systems, type_map=type_map)
        for i, (artifact, atoms) in enumerate(atoms_list):
            data_file = os.path.join(base_dir, f'{i:06d}.lammps.data')
            ase.io.write(data_file, atoms, format='lammps-data', specorder=type_map)  # type: ignore
            data_files.append({
                'url': data_file,
                'attrs': artifact['attrs'],
            })
        return data_files

    return (
        artifacts_to_ase_atoms,
        ase_atoms_to_cp2k_input_data,
        convert_to_lammps_input_data,
        DataFormat,
        get_data_format,
    )

(
    artifacts_to_ase_atoms,
    ase_atoms_to_cp2k_input_data,
    convert_to_lammps_input_data,
    DataFormat,
    get_data_format,
) = __export_remote_functions()



from asaplib.data.xyz import ASAPXYZ
from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import dump_json, flush_stdio

from typing import List, Optional, Tuple, Dict
from io import StringIO
from pydantic import BaseModel
from dataclasses import dataclass
import pandas as pd
from tabulate import tabulate
from itertools import groupby
from functools import lru_cache

import ase.io
import os

from .data import get_data_format, DataFormat, artifacts_to_ase_atoms
from .iface import ICllSelectorOutput, BaseCllContext
from .constant import LAMMPS_DUMP_DIR, LAMMPS_DUMP_SUFFIX, DEFAULT_ASAP_SOAP_DESC, DEFAULT_ASAP_PCA_REDUCER
from .asap import get_descriptor, reduce_dimension, get_trainer, get_cluster


logger = get_logger(__name__)


class CllModelDeviSelectorInputConfig(BaseModel):

    class AsapOptions(BaseModel):
        disable: bool = False
        limit_per_cluster: int = 1
        """
        limit the number of structures to be selected from the same cluster
        """
        sort_by_ssw_energy: bool = False
        """
        sorted the structures by ssw_energy in each cluster
        """
        descriptor: dict = {'soap': { **DEFAULT_ASAP_SOAP_DESC, 'preset': 'minimal'}}
        dim_reducer: dict = {'pca': DEFAULT_ASAP_PCA_REDUCER}
        cluster: dict = {'dbscan': {}}

    f_trust_lo: float = 0.
    """
    the lower bound of model_devi score to select the structure for labeling
    """
    f_trust_hi: float = 65535.
    """
    the upper bound of model_devi score to select the structure for labeling
    """
    new_explore_system_q: float = 0.25
    """
    the quantile of model_devi score to select the structure for next round of exploration
    """
    asap_options: Optional[AsapOptions]
    """
    options for ASAP to further select candidates
    """
    screening_fn: Optional[str] = None
    """
    the function to screen the candidates, e.g
    "lambda x: x['ssw_energy'] < -1000"
    """


@dataclass
class CllModelDevSelectorContext(BaseCllContext):
    ...


@dataclass
class CllModelDeviSelectorOutput(ICllSelectorOutput):
    candidates: List[Artifact]
    passing_rate: float
    new_explore_systems: List[Artifact]

    def get_model_devi_dataset(self):
        return self.candidates

    def get_passing_rate(self) -> float:
        return self.passing_rate

    def get_new_explore_systems(self) -> List[Artifact]:
        return self.new_explore_systems


@dataclass
class CllModelDeviSelectorInput:
    config: CllModelDeviSelectorInputConfig
    model_devi_data: List[Artifact]
    model_devi_file: str
    type_map: List[str]

    def set_model_devi_dataset(self, data: List[Artifact]):
        self.model_devi_data = data


async def cll_model_devi_selector(input: CllModelDeviSelectorInput, ctx: CllModelDevSelectorContext):
    executor = ctx.resource_manager.default_executor
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    executor.mkdir(work_dir)

    f_trust_lo = input.config.f_trust_lo
    f_trust_hi = input.config.f_trust_hi

    results = executor.run_python_fn(bulk_select_structures_by_model_devi)(
        model_devi_outputs=[a.to_dict() for a in input.model_devi_data],
        model_devi_file=input.model_devi_file,
        f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
        type_map=input.type_map, work_dir=work_dir,
        new_explore_system_q=input.config.new_explore_system_q,
        screening_fn=input.config.screening_fn,
    )

    candidates = [ result['decent'] for result, _ in results if 'decent' in result ]
    new_systems = [ result['next'] for result, _ in results if 'next' in result ]
    stats = [ stats for _, stats in results ]

    # group next_structures by `attrs.source` and keep the first one
    # so that the total number of explore structures will be the same as the original one
    new_systems = [next(group) for _source, group in groupby(
        new_systems, key=lambda s: s['attrs']['source'])]

    # write model_deviation stats report
    # TODO: refactor into a function
    def _get_row(stat: dict):
        url = stat['src']
        url = f'...{url[-30:]}' if len(url) > 30 else url  # wrap url if it is too long
        total, good, decent, poor = stat['total'], stat['good'], stat['decent'], stat['poor']
        return [
            url, total, good, decent, poor,
            f'{good / total * 100:.2f}%', f'{decent / total * 100:.2f}%', f'{poor / total * 100:.2f}%',
        ]
    headers = ['file', 'total', 'good', 'decent', 'poor', 'good%', 'decent%', 'poor%']
    table = [ _get_row(stat) for stat in stats]
    total = sum(row[1] for row in table)
    total_good = sum(row[2] for row in table)

    stats_report = tabulate(table, headers=headers, tablefmt='tsv')
    logger.info('stats report: \n%s\n', stats_report)
    executor.dump_text(stats_report, os.path.join(work_dir, 'stats.tsv'))

    # further select candidates by ASAP
    if input.config.asap_options and not input.config.asap_options.disable:
        asap_options = input.config.asap_options
        candidates = executor.run_python_fn(bulk_select_distinct_structures)(
            candidates=candidates,
            descriptor_opt=asap_options.descriptor,
            dim_reducer_opt=asap_options.dim_reducer,
            cluster_opt=asap_options.cluster,
            type_map=input.type_map,
            work_dir=work_dir,
            limit_per_cluster=asap_options.limit_per_cluster,
            sort_by_energy=asap_options.sort_by_ssw_energy,
        )

    return CllModelDeviSelectorOutput(
        candidates=[Artifact.of(**a) for a in candidates],
        new_explore_systems=[Artifact.of(**a) for a in new_systems],
        passing_rate=total_good / total,
    )


def __export_remote_functions():

    def bulk_select_structures_by_model_devi(model_devi_outputs: List[ArtifactDict],
                                             model_devi_file: str,
                                             f_trust_lo: float,
                                             f_trust_hi: float,
                                             new_explore_system_q: float,
                                             type_map: List[str],
                                             work_dir: str,
                                             workers: int = 4,
                                             screening_fn: Optional[str] = None,
                                             ) -> List[Tuple[Dict[str, ArtifactDict], dict]]:
        import joblib
        return joblib.Parallel(n_jobs=workers)(
            joblib.delayed(select_structures_by_model_devi)(
                model_devi_output=output,
                model_devi_file=model_devi_file,
                f_trust_lo=f_trust_lo, f_trust_hi=f_trust_hi,
                type_map=type_map,
                work_dir=os.path.join(work_dir, 'model_devi', f'{i:06}'),
                new_explore_system_q=new_explore_system_q,
                screening_fn=screening_fn,
            )
            for i, output in enumerate(model_devi_outputs)
        )  # type: ignore


    def select_structures_by_model_devi(model_devi_output: ArtifactDict,
                                        model_devi_file: str,
                                        f_trust_lo: float,
                                        f_trust_hi: float,
                                        type_map: List[str],
                                        work_dir: str,
                                        new_explore_system_q: float,
                                        screening_fn: Optional[str] = None,
                                        ) -> Tuple[Dict[str, ArtifactDict], dict]:
        """
        analysis the model_devi output of explore stage and select candidates

        :param next_explore_system_q: the quantile of model_devi score to select the structure for next round of exploration
        """
        os.makedirs(work_dir, exist_ok=True)
        dump_json(model_devi_output, os.path.join(work_dir, 'model_devi_output.debug.json'))

        model_devi_dir = model_devi_output['url']
        model_devi_file = model_devi_output['attrs'].pop('model_devi_file', model_devi_file)

        force_col = 'max_devi_f'
        print(f'criteria: {f_trust_lo} <= {force_col} < {f_trust_hi}')

        # get path of model_devi file
        data_format = get_data_format(model_devi_output)  # type: ignore
        if data_format in (DataFormat.LAMMPS_OUTPUT_DIR, DataFormat.LASP_LAMMPS_OUT_DIR):
            model_devi_file = os.path.join(model_devi_dir, model_devi_file)
        else:
            raise ValueError('unknown model_devi_data types')
        logger.info('start to analysis file: %s', model_devi_file)

        # load model_devi data
        with open(model_devi_file, 'r') as f:
            text = f.read()
        df = pd.read_csv(StringIO(text.lstrip('#')), delim_whitespace=True)
        # layout:
        #        step  max_devi_v  min_devi_v  avg_devi_v  max_devi_f  min_devi_f  avg_devi_f
        # 0        0    0.006793    0.000672    0.003490    0.143317    0.005612    0.026106
        # 1      100    0.006987    0.000550    0.003952    0.128178    0.006042    0.022608

        # load structures
        atoms_list = []
        if data_format == DataFormat.LAMMPS_OUTPUT_DIR:
            lammps_dump_dir = model_devi_output['attrs'].pop('lammps_dump_dir', LAMMPS_DUMP_DIR)
            for frame_id in df.step:
                dump_file = os.path.join(model_devi_dir, lammps_dump_dir, f'{frame_id}{LAMMPS_DUMP_SUFFIX}')
                atoms_list += ase.io.read(dump_file, ':', format='lammps-dump-text', specorder=type_map)
        elif data_format == DataFormat.LASP_LAMMPS_OUT_DIR:
            structures_file = os.path.join(model_devi_dir, 'structures.xyz')
            atoms_list += ase.io.read(structures_file, ':', format='extxyz')
        else:
            raise ValueError('unknown model_devi_data types')

        # screening structure before model_devi analysis
        if screening_fn is not None:

            if 'ssw_energy' in atoms_list[0].info:
                s_ssw_energy = pd.Series(map(lambda atoms: atoms.info['ssw_energy'], atoms_list))  # type: ignore

                # the following ssw_* methods are for the screening_fn
                # so don't need to worry about the unused warning
                ssw_energy_max = s_ssw_energy.max()
                ssw_energy_min = s_ssw_energy.min()
                def ssw_energy_quantile(q):
                    # the quantile will be evaluated every time, which is not efficient
                    # so here we cache the result, carefully
                    return lru_cache()(s_ssw_energy.quantile)(q)
                ssw_enenrgy_quantile = ssw_energy_quantile

            # return the df row whose atoms pass the screening_fn
            _screening_fn = eval(screening_fn, locals())  # str to function
            df = df[[ _screening_fn(atoms) for atoms in atoms_list ]]


        # evaluate new found structures by their model_devi score in 3 levels: good, decent, poor
        good_df   = df[df[force_col] < f_trust_lo]
        decent_df = df[(df[force_col] >= f_trust_lo) & (df[force_col] < f_trust_hi)]
        poor_df   = df[df[force_col] >= f_trust_hi]

        # select the last frame from df whose model_devi score is less than the quantile
        # as the initial structure for next round of exploration to replace the original one
        # NOTE: equal is essential to ensure the existence of next structure
        # NOTE: select the last frame can increase the diversity of structures
        next_df = df[df[force_col] <= df[force_col].quantile(new_explore_system_q)].tail(1)

        stats = {
            'src': model_devi_file,
            'total': len(df),
            'good': len(good_df),
            'decent': len(decent_df),
            'poor': len(poor_df),
        }


        result: Dict[str, ArtifactDict] = {}

        # dump structures to different files
        # TODO: refactor the following repeating code
        if len(good_df) > 0:
            good_file = os.path.join(work_dir, 'good.xyz')
            ase.io.write(good_file, [atoms_list[_i] for _i in good_df.index], format='extxyz')
            result['good'] = {'url': good_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                              'attrs': {**model_devi_output['attrs']}}
        if len(poor_df) > 0:
            poor_file = os.path.join(work_dir, 'poor.xyz')
            ase.io.write(poor_file, [atoms_list[_i] for _i in poor_df.index], format='extxyz')
            result['poor'] = {'url': poor_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                              'attrs': {**model_devi_output['attrs']}}
        if len(decent_df) > 0:
            decent_file = os.path.join(work_dir, 'decent.xyz')
            ase.io.write(decent_file, [atoms_list[_i] for _i in decent_df.index], format='extxyz')
            result['decent'] = {'url': decent_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                                'attrs': {**model_devi_output['attrs']}}
        if len(next_df) > 0:
            next_file = os.path.join(work_dir, 'next.xyz')
            ase.io.write(next_file, [atoms_list[_i] for _i in next_df.index], format='extxyz')
            result['next'] = {'url': next_file, 'format': DataFormat.EXTXYZ,  # type: ignore
                              'attrs': {**model_devi_output['attrs']}}

        dump_json([result, stats, list(decent_df.index), list(next_df.index)], os.path.join(work_dir, 'result.debug.json'))

        return result, stats


    def bulk_select_distinct_structures(candidates: List[ArtifactDict],
                                        descriptor_opt: dict,
                                        dim_reducer_opt: dict,
                                        cluster_opt: dict,
                                        type_map: List[str],
                                        work_dir: str,
                                        limit_per_cluster: int = -1,
                                        sort_by_energy: bool = False,
                                        workers: int = 4,
                                        ) -> List[ArtifactDict]:
        inputs = []
        for i, (ancestor_key, candidate_group) in enumerate(groupby(candidates, key=lambda c: c['attrs']['ancestor'])):
            candidate_group = list(candidate_group)
            inputs.append((candidate_group, candidate_group[0]['attrs']))

        import joblib
        return joblib.Parallel(n_jobs=workers)(
            joblib.delayed(select_distinct_structures)(
                candidates=group,
                attrs=attrs,
                descriptor_opt=descriptor_opt,
                dim_reducer_opt=dim_reducer_opt,
                cluster_opt=cluster_opt,
                type_map=type_map,
                work_dir=os.path.join(work_dir, 'asap', f'{i:06}'),
                limit_per_cluster=limit_per_cluster,
                sort_by_energy=sort_by_energy,
            ) for i, (group, attrs) in enumerate(inputs)
        )  # type: ignore


    def select_distinct_structures(candidates: List[ArtifactDict],
                                   attrs: dict,
                                   descriptor_opt: dict,
                                   dim_reducer_opt: dict,
                                   cluster_opt: dict,
                                   type_map: List[str],
                                   work_dir: str,
                                   limit_per_cluster: int = -1,
                                   sort_by_energy: bool = False,
                                   ):
        os.makedirs(work_dir, exist_ok=True)

        dump_json(attrs, os.path.join(work_dir, 'attrs.debug.json'))
        # load structures and save it to a tmp file for ASAP to load
        atoms_list = [atoms for _, atoms in artifacts_to_ase_atoms(candidates, type_map=type_map)]


        if len(atoms_list) < 10:
            # FIXME: there are a lot of potential issue when the number of atoms is small
            # the root cause is in asaplib, which I guess has not been tested with small dataset
            selected_atoms_list = atoms_list
        else:
            if sort_by_energy and 'ssw_energy' in atoms_list[0].info:
                atoms_list = sorted(atoms_list, key=lambda atoms: atoms.info['ssw_energy'])

            # use asaplib to group structures
            # load structures to ASAP
            tmp_structures_file = os.path.join(work_dir, '.tmp-structures.xyz')
            ase.io.write(tmp_structures_file, atoms_list, format='extxyz')
            asapxyz = ASAPXYZ(tmp_structures_file)

            # group structures
            asap_path_prefix = os.path.join(work_dir, 'asap')
            descriptors, _ = get_descriptor(asapxyz, descriptor_opt, path_prefix=asap_path_prefix)
            reduced_descriptors = reduce_dimension(descriptors, dim_reducer_opt)
            trainer = get_trainer(reduced_descriptors, cluster_opt)
            cluster_labels = get_cluster(asapxyz, reduced_descriptors, trainer, path_prefix=asap_path_prefix)

            # dump_json(cluster_labels, os.path.join(work_dir, 'cluster.debug.json'))

            # select frame from each cluster
            if limit_per_cluster <= 0: # unlimit
                selected_atoms_list = atoms_list  # do nothing
            else:
                selected_frames = []
                for frames in cluster_labels.values():
                    if len(frames) < limit_per_cluster:
                        selected_frames += list(frames)
                    else:
                        selected_frames += list(frames[:limit_per_cluster])
                selected_atoms_list = [atoms_list[i] for i in selected_frames]

        # write selected structures to file
        distinct_structures_file = os.path.join(work_dir,  'distinct_structures.xyz')
        ase.io.write(distinct_structures_file, selected_atoms_list, format='extxyz')

        output = {
            'url': distinct_structures_file,
            'format': DataFormat.EXTXYZ,
            'attrs': attrs,
        }
        dump_json(output, os.path.join(work_dir, 'output.debug.json'))
        flush_stdio()  # flush joblib stdio buffer
        return output

    return (
        bulk_select_structures_by_model_devi,
        bulk_select_distinct_structures,
    )

(
    bulk_select_structures_by_model_devi,
    bulk_select_distinct_structures,
) = __export_remote_functions()



# Suppress numba warning of umap.
# Ref: https://github.com/lmcinnes/umap/issues/252
from numba.core.errors import NumbaDeprecationWarning
import warnings
warnings.filterwarnings("ignore", category=NumbaDeprecationWarning)

from asaplib.data.xyz import ASAPXYZ
from asaplib.cli.func_asap import cluster_process, set_reducer
from asaplib.hypers.hyper_soap import universal_soap_hyper
from asaplib.hypers.hyper_acsf import universal_acsf_hyper
from asaplib.reducedim.dim_reducer import Dimension_Reducers
from asaplib.cluster.ml_cluster_fit import LAIO_DB, sklearn_DB

from scipy.spatial.distance import cdist

import numpy as np


def __export_remote_functions():

    def get_descriptor(asapxyz: ASAPXYZ, setting: dict, keep_atomic=False, n_process=4, path_prefix='./asap-descriptor'):

        descriptor_spec = get_descriptor_spec(asapxyz, setting)
        asapxyz.compute_global_descriptors(
            desc_spec_dict = descriptor_spec,
            keep_atomic = keep_atomic,
            n_process = n_process,
        )
        asapxyz.save_state(path_prefix)

        global_descriptors = asapxyz.fetch_computed_descriptors(list(descriptor_spec.keys()))
        atomic_descriptors = asapxyz.fetch_computed_atomic_descriptors(list(descriptor_spec.keys())) if keep_atomic else None
        return global_descriptors, atomic_descriptors


    def get_descriptor_spec(asapxyz: ASAPXYZ, setting: dict):
        descriptor_type, params = None, None
        for key in ['soap', 'acsf', 'cm']:
            params = setting.get(key, None)
            if params is not None:
                descriptor_type = key
                break
        else:
            raise ValueError('Unknown descriptor type')


        if descriptor_type == 'cm':
            return {'cm': {'type': 'CM'}}

        reducer_spec = dict(set_reducer(
            params['reducer_type'],
            params['element_wise'],
            params['zeta'],
        ))

        if descriptor_type == 'acsf':
            atomic_descriptor_spec = get_acsf_atomic_descriptor_spec(asapxyz, params)
        elif descriptor_type == 'soap':
            atomic_descriptor_spec = get_soap_atomic_descriptor_spec(asapxyz, params)

        descriptor_spec = {}
        for k, v in atomic_descriptor_spec.items():  # type: ignore
            descriptor_spec[k] = {
                'atomic_descriptor': {k: v},
                'reducer_function': reducer_spec,
            }
        return descriptor_spec


    def get_soap_atomic_descriptor_spec(asapxyz: ASAPXYZ, params: dict):
        preset = params['preset']
        if preset is None:
           atomic_descriptor_spec = {
                'soap1': {
                    'type': 'SOAP',
                    'cutoff': params['r_cut'],
                    'n': params['n_max'],
                    'l': params['l_max'],
                    'atom_gaussian_width': params['sigma'],
                }
            }
        else:
            atomic_descriptor_spec = universal_soap_hyper(asapxyz.get_global_species(), preset, dump=False)

        for k in atomic_descriptor_spec.keys():
            atomic_descriptor_spec[k]['rbf'] = params['rbf']
            atomic_descriptor_spec[k]['crossover'] = params['crossover']

        return atomic_descriptor_spec


    def get_acsf_atomic_descriptor_spec(asapxyz: ASAPXYZ, params: dict):
        preset = params['preset']
        facsf_param = params['r_cut'] if preset is None else preset
        return universal_acsf_hyper(asapxyz.get_global_species(), facsf_param, dump=False)


    def reduce_dimension(descriptors: np.ndarray, setting: dict):
        reducer = Dimension_Reducers(setting)
        reduced_descriptors = reducer.fit_transform(descriptors)
        return reduced_descriptors


    def get_dbscan_trainer(descriptors: np.ndarray, metric='euclidean', eps=None, min_samples=2, eval_sample=50):
        if eps is None:
            n = len(descriptors)
            samples = descriptors[np.random.choice(n, min(n, eval_sample), replace=False)]
            # FIXME: the method to estimate eps is strange
            # FIXME: this will be broken when len of descriptors small
            eps = np.percentile(cdist(samples, descriptors, metric), min(100 * 10. / n, 99))  # type: ignore
        return sklearn_DB(eps, min_samples, metrictype=metric)


    def get_laio_db_trainer(**kwargs):
        return LAIO_DB(**kwargs)


    def get_trainer(descriptor: np.ndarray, setting: dict):
        dbscan_setting = setting.get('dbscan', None)
        if dbscan_setting is not None:
            return get_dbscan_trainer(descriptor, **dbscan_setting)
        laiodb_setting = setting.get('laiodb', None)
        if laiodb_setting is not None:
            return get_laio_db_trainer(**laiodb_setting)
        raise ValueError('Unknown trainer type')


    def get_cluster(asapxyz: ASAPXYZ, descriptors: np.ndarray, trainer, path_prefix='./asap-cluster'):
        options = {
            'prefix': path_prefix,
            'savexyz': False,
            'savetxt': False,
        }
        labels: np.ndarray = cluster_process(asapxyz, trainer, descriptors, options)  # type: ignore

        groups = {}
        index = np.arange(len(labels))
        for label in set(labels):
            groups[label] = index[labels == label]
        return groups

    return (
        get_descriptor,
        reduce_dimension,
        get_trainer,
        get_cluster,
    )


(
    get_descriptor,
    reduce_dimension,
    get_trainer,
    get_cluster,
) = __export_remote_functions()


import ase.io
import fire
import numpy as np
import matplotlib.pyplot as plt

from ruamel.yaml import YAML
from ruamel.yaml.comments import TaggedScalar

from ai2_kit.core.log import get_logger
from ai2_kit.core.util import parse_path_list, nested_set
from ai2_kit.domain.cp2k import dump_coord_n_cell
from ai2_kit.domain.lammps import get_ensemble
from ai2_kit import res

from typing import Optional, Literal, List
from ase import Atoms, Atom
from string import Template
import random
import re
import os
import json
import io


logger = get_logger(__name__)

AI2CAT_RES_DIR = os.path.join(res.DIR_PATH, 'catalysis')

DEEPMD_DEFAULT_TEMPLATE = os.path.join(AI2CAT_RES_DIR, 'deepmd.json')
MLP_TRAINING_TEMPLATE   = os.path.join(AI2CAT_RES_DIR, 'mlp-training.yml')
CP2K_DEFAULT_TEMPLATE   = os.path.join(AI2CAT_RES_DIR, 'cp2k.inp')

CP2K_SCF_SEMICONDUCTOR = """\
        # CONFIGURATION FOR SEMICONDUCTOR
        &SCF
            SCF_GUESS RESTART
            EPS_SCF 3e-07
            MAX_SCF 50
            &OUTER_SCF
                EPS_SCF 3e-07
                MAX_SCF 20
            &END OUTER_SCF
            &OT
                MINIMIZER CG
                PRECONDITIONER FULL_SINGLE_INVERSE
                ENERGY_GAP 0.1
            &END OT
        &END SCF
        # END CONFIGURATION FOR SEMICONDUCTOR"""

CP2K_SCF_METAL = """\
        # CONFIGURATION FOR METAL
        &SCF
            SCF_GUESS RESTART
            EPS_SCF 3e-07
            MAX_SCF 500
            ADDED_MOS 500
            CHOLESKY INVERSE
            &SMEAR
                METHOD FERMI_DIRAC
                ELECTRONIC_TEMPERATURE [K] 300
            &END SMEAR
            &DIAGONALIZATION
                ALGORITHM STANDARD
            &END DIAGONALIZATION
            &MIXING
                METHOD BROYDEN_MIXING
                ALPHA 0.3
                BETA 1.5
                NBROYDEN 14
            &END MIXING
        &END SCF
        # END CONFIGURATION FOR METAL"""

CP2K_MOTION_TEMPLATE = """\
&MOTION
  &MD
    ENSEMBLE NVT
    STEPS       $steps
    TIMESTEP    $timestep
    TEMPERATURE $temp
    &THERMOSTAT
       TYPE CSVR
       REGION MASSIVE
       &CSVR
          TIMECON [fs] 100.0
       &END
    &END
  &END MD
  &PRINT
   &TRAJECTORY
     &EACH
       MD 1
     &END EACH
   &END TRAJECTORY
   &VELOCITIES
     &EACH
       MD 1
     &END EACH
   &END VELOCITIES
   &FORCES
     &EACH
       MD 1
     &END EACH
   &END FORCES
   &RESTART_HISTORY
     &EACH
       MD 1000
     &END EACH
   &END RESTART_HISTORY
   &RESTART
     BACKUP_COPIES 3
     &EACH
       MD 1
     &END EACH
   &END RESTART
  &END PRINT
&END MOTION"""

CP2K_SCF_TABLE = {
    'metal': CP2K_SCF_METAL,
    'semi': CP2K_SCF_SEMICONDUCTOR,
}

CP2K_ACCURACY_TABLE = {
    'high': {'cutoff': 1000, 'rel_cutoff': 90 },
    'medium': {'cutoff': 800, 'rel_cutoff': 70 },
    'low': {'cutoff': 600, 'rel_cutoff': 50 },
}


class ConfigBuilder:

    def __init__(self):
        self._atoms: Optional[Atoms] = None

    def load_system(self, file: str, **kwargs):
        """
        Loading system to memory using ASE
        """
        assert self._atoms is None, 'atoms is already loaded'
        kwargs.setdefault('index', 0)
        self._atoms = ase.io.read(file, **kwargs)  # type: ignore
        return self

    def gen_mlp_training_input(self,
                               out_dir: str = 'out',
                               train_data: Optional[List[str]] = None,
                               explore_data: Optional[List[str]] = None,
                               artifacts: Optional[List[dict]] = None,
                               template_file: str = MLP_TRAINING_TEMPLATE):
        train_data = train_data or []
        explore_data = explore_data or []
        artifacts = artifacts or []

        # Read yaml file as text so that the comments are preserved
        with open(template_file, 'r') as fp:
            text = fp.read()
        # Generate the type_map and mass_map automatically
        assert self._atoms is not None, 'atoms must be loaded first'
        type_map, mass_map = get_type_map(self._atoms)

        out_data = Template(text).substitute(
            type_map=json.dumps(type_map),
            mass_map=json.dumps(mass_map),
            train_data=json.dumps(train_data),
            explore_data=json.dumps(explore_data),
            artifacts=dump_artifacts(artifacts),
        )
        os.makedirs(out_dir, exist_ok=True)
        mlp_training_input_path = os.path.join(out_dir, 'training.yml')
        with open(mlp_training_input_path, 'w', encoding='utf-8') as fp:
            fp.write(out_data)

    def gen_deepmd_input(self,
                         out_dir: str = 'out',
                         steps: int = 10000,
                         template_file: str = DEEPMD_DEFAULT_TEMPLATE):
        with open(template_file, 'r') as fp:
            data = json.load(fp)
        os.makedirs(out_dir, exist_ok=True)

        data['training']['numb_steps'] = steps
        deepmd_input_path = os.path.join(out_dir, 'deepmd.json')
        with open(deepmd_input_path, 'w', encoding='utf-8') as fp:
            json.dump(data, fp, indent=4)

    def gen_plumed_input(self, out_dir: str = 'out'):
        """
        Generate Plumed input files
        Args:
            out_dir: output directory, plumed.inp will be generated in this directory
        """
        assert self._atoms is not None, 'atoms must be loaded first'
        plumed_input = [ 'UNITS LENGTH=A' ]
        # define atoms groups by element, e.g:
        # Ag: GROUP ATOMS=1,2,3,4,5,6,7,8,9,10
        # O: GROUP ATOMS=11,12,13,14,15,16,17,18,19,20
        element2ids = {}
        for i, element in enumerate(self._atoms.get_chemical_symbols(), start=1):
            element2ids.setdefault(element, []).append(i)
        for element, ids in element2ids.items():
            plumed_input.append(f'{element}: GROUP ATOMS={",".join(map(str, ids))}')

        # define reaction coordinate
        plumed_input.extend([
            '#define more groups if needed',
            '',
            '# define reaction coordinates, e.g. CV1, CV2, ...',
            '# you may define as many as you want',
            'CV1:',
            '',
            '# define sampling method: metadynamics',
            'metad: METAD ARG=CV1 SIGMA=0.1 HEIGHT=5 PACE=100 FILE=HILLS',
            '# define more commands if you need',
            '',
            '# print CVs',
            'PRINT STRIDE=10 ARG=CV1,metad.bias FILE=COLVAR',
        ])
        os.makedirs(out_dir, exist_ok=True)
        plumed_input_path = os.path.join(out_dir, 'plumed.inp')
        with open(plumed_input_path, 'w', encoding='utf-8') as fp:
            fp.write('\n'.join(plumed_input))

    def gen_lammps_input(self, out_dir='./out', **kwargs):
        assert self._atoms is not None, 'atoms must be loaded first'
        kwargs = {
            'nsteps': 20000,
            'stepsize': 0.0005,
            'temp': 330,
            'sample_freq': 100,
            'pres': -1,
            'tau_t': 0.1,
            'tau_p': 0.5,
            'time_const': 0.1,
            'model_devi_out': 'model_devi.out',
            'dump_out': 'traj.lammpstrj',
            'energy_out': 'energy.log',
            'data_file': 'lammps.dat',
            'lammps_file': 'lammps.inp',
            'plumed_file': 'plumed.inp',
            'plumed_out': 'plumed.out',
            ** kwargs,
        }

        template_file  = kwargs.pop('template_file', os.path.join(AI2CAT_RES_DIR, 'lammps-post.inp'))
        dp_models = parse_path_list(kwargs.pop('dp_models'), to_abs=True)
        ensemble = kwargs.pop('ensemble')

        type_map, mass_map = get_type_map(self._atoms)
        ensemble_config = '\n'.join([
            '''velocity all create ${TEMP} %d''' % (random.randrange(10^6 - 1) + 1),
            get_ensemble(ensemble),
        ])

        template_vars = {
            **kwargs,
            'type_map': ' '.join(type_map),
            'mass_map': '\n'.join([f'mass {i} {m}' for i, m in enumerate(mass_map, start=1)]),
            'dp_models': ' '.join(dp_models),
            'ensemble_config': ensemble_config,
        }

        class LammpsTemplate(Template):
            delimiter = '$$'

        with open(template_file, 'r') as fp:
            template = LammpsTemplate(fp.read())
        lammps_input = template.substitute(**template_vars)

        # write lammps input and data to output dir
        os.makedirs(out_dir, exist_ok=True)
        lammps_file_path = os.path.join(out_dir, kwargs['lammps_file'])
        with open(lammps_file_path, 'w', encoding='utf8') as fp:
            fp.write(lammps_input)
        lammps_data_path = os.path.join(out_dir, kwargs['data_file'])
        ase.io.write(lammps_data_path, self._atoms, format='lammps-data', specorder=type_map)  # type: ignore


    def gen_cp2k_input(self,
                       out_dir: str = 'out',
                       basic_set_file: str = 'BASIS_MOLOPT',
                       potential_file: str = 'GTH_POTENTIALS',
                       style: Literal['metal', 'semi'] = 'metal',
                       accuracy: Literal['high', 'medium', 'low'] = 'medium',
                       aimd: bool = False,
                       temp: float = 330.,
                       steps: int = 1000,
                       timestep: float = 0.5,
                       parameter_file: str = 'dftd3.dat',
                       template_file: str = CP2K_DEFAULT_TEMPLATE,):
        """
        Generate CP2K input files

        You should call `load_system` first to load system into memory.
        And you may also need to ensure the basic set and potential files
        you want to use are available in CP2K_DATA_DIR,
        or else you have to specify the full path instead of their file name.

        Args:
            basic_set_file: basic set file, can be path or file in CP2K_DATA_DIR
            potential_file: potential file, can be path or file in CP2K_DATA_DIR
            out_dir: output directory, cp2k.inp and coord_n_cell.inc will be generated in this directory
            style: 'metal' or 'semi'
            accuracy: 'high', 'medium' or 'low'
            aimd: whether to run AIMD
            temp: temperature for AIMD
            steps: steps for AIMD
            timestep: timestep for AIMD
            parameter_file: parameter file, can be path or file in CP2K_DATA_DIR
            template_file: template file, no need to change in most cases
        """

        assert self._atoms is not None, 'atoms must be loaded first'
        # get available basic set and potential
        with open(find_cp2k_data_file(basic_set_file), 'r') as fp:
            basic_set_table = parse_cp2k_data_file(fp)
        with open(find_cp2k_data_file(potential_file), 'r') as fp:
            potential_table = parse_cp2k_data_file(fp)
        with open(template_file, 'r') as fp:
            template = fp.read()
        # build kinds
        elements = set(self._atoms.get_chemical_symbols())

        kinds = ''
        total_ve = 0  # Valence Electron

        for element in elements:
            basic_set_all = basic_set_table.get(element)
            basic_set = basic_set_all[-1]
            logger.info(f'BASIC_SET {basic_set} is selected for {element}')

            potential_all = potential_table.get(element)
            potential = next(p for p in potential_all if 'PBE' in p)
            logger.info(f'POTENTIAL {potential} is selected for {element}')
            kinds += '\n'.join([
                f'    &KIND {element}',
                f'        BASIS_SET {basic_set}',
                f'        # All available BASIS_SET:',
                f'        # {" ".join(basic_set_all)}',
                f'        POTENTIAL {potential}',
                f'        # All available POTENTIAL:',
                f'        # {" ".join(potential_all)}',
                f'    &END KIND',
                '', # This empty line is required
            ])
            # get ve from the last number of potential
            # for example, if the potential is GTH-OLYP-q6
            # then the ve is 6
            ve = int(re.match(r'.*?(\d+)$', potential).group(1))
            total_ve += ve * self._atoms.get_chemical_symbols().count(element)
        logger.info("total valence electron of the system: %d" % total_ve)

        motion = ''
        if aimd:
            motion = Template(CP2K_MOTION_TEMPLATE).substitute(
                steps=steps,
                timestep=timestep,
                temp=temp,
            )

        template_vars = {
            'run_type': 'MD' if aimd else 'ENERGY_FORCE',
            'basic_set_file': os.path.abspath(basic_set_file),
            'potential_file': os.path.abspath(potential_file),
            'parameter_file': parameter_file,
            'uks': 'T' if total_ve % 2 else 'F',
            'kinds': kinds,
            'motion': motion,
            'scf': CP2K_SCF_TABLE[style],
            **CP2K_ACCURACY_TABLE[accuracy],
        }

        cp2k_input = Template(template).substitute(**template_vars)
        os.makedirs(out_dir, exist_ok=True)
        cp2k_input_path = os.path.join(out_dir, 'cp2k.inp')
        with open(cp2k_input_path, 'w') as fp:
            fp.write(cp2k_input)
        logger.info(f'CP2K input file is generated at {cp2k_input_path}')

        coord_n_cell_path = os.path.join(out_dir, 'coord_n_cell.inc')
        with open(coord_n_cell_path, 'w') as fp:
            dump_coord_n_cell(fp, self._atoms)
        logger.info(f'coord_n_cell.inc is generated at {coord_n_cell_path}')

    def get_plumed_group(self):
        """
        Get auto generated plumed group
        """
        assert self._atoms is not None, 'atoms must be loaded first'
        plumed_input = []
        # define atoms groups by element, e.g:
        # Ag: GROUP ATOMS=1,2,3,4,5,6,7,8,9,10
        # O: GROUP ATOMS=11,12,13,14,15,16,17,18,19,20
        element2ids = {}
        for i, element in enumerate(self._atoms.get_chemical_symbols(), start=1):
            element2ids.setdefault(element, []).append(i)
        for element, ids in element2ids.items():
            plumed_input.append(f'{element}: GROUP ATOMS={",".join(map(str, ids))}')
        return '\n'.join(plumed_input)


def get_type_map(atoms: Atoms):
    type_map = sorted(set(atoms.get_chemical_symbols()))  # sorted to ensure order
    mass_map = [Atom(symbol).mass for symbol in type_map]
    return type_map, mass_map


def find_cp2k_data_file(file: str):
    """
    find CP2K data file in CP2K_DATA_DIR
    """
    if not os.path.exists(file):
        cp2k_data_dir = os.environ.get('CP2K_DATA_DIR')
        if cp2k_data_dir is None:
            raise FileNotFoundError(f'Cannot find {file}')
        file = os.path.join(cp2k_data_dir, file)
        if not os.path.exists(file):
            raise FileNotFoundError(f'Cannot find {file}')
    return file


def parse_cp2k_data_file(fp):
    """
    get all available basic set and potential from CP2K data file
    """
    parsed = {}
    for line in fp:
        line: str = line.strip()
        if line.startswith('#'):
            continue # skip comment
        tokens = line.split()
        if 0 == len(tokens):
            continue  # skip empty line
        if re.match(r'^[A-Z][a-z]*$', tokens[0]):
            parsed.setdefault(tokens[0], []).append(tokens[1])
    return parsed


def dump_artifacts(artifacts: List[dict]) -> str:
    out = {}
    for artifact in artifacts:
        key = artifact['key']
        out[key] = {}
        out[key]['url'] = artifact['url']
        attrs = {}
        out[key]['attrs'] = attrs

        if 'cp2k_file' in artifact:
            cp2k_file = os.path.abspath(artifact['cp2k_file'])
            nested_set(attrs, ['cp2k', 'input_template'], TaggedScalar(value=cp2k_file, tag='!load_text'))
        if 'plumed_file' in artifact:
            plumed_file = os.path.abspath(artifact['plumed_file'])
            nested_set(attrs, ['lammps', 'plumed_config', ], TaggedScalar(value=plumed_file, tag='!load_text'))

    yaml = YAML()
    yaml.default_flow_style = False
    buf = io.BytesIO()
    yaml.dump({'artifacts': out}, buf)
    return buf.getvalue().decode('utf-8')


def inspect_lammps_output(lammps_dir: str, save_to: Optional[str]=None, fig_ax = None):
    model_devi_file = os.path.join(lammps_dir, 'model_devi.out')
    colvar_file = os.path.join(lammps_dir, 'COLVAR')
    lammps_input_file = os.path.join(lammps_dir, 'lammps.input')

    # read TEMP, N_STEPS lammps input file
    with open(lammps_input_file, 'r') as fp:
        for line in fp:
            line = line.strip()
            if line.startswith('variable'):
                tokens = line.split()
                if tokens[1] == 'TEMP':
                    temp = float(tokens[3])

    # draw colvar
    colvar = np.loadtxt(colvar_file, skiprows=1)
    # read column names from the first line of colvar file
    with open(colvar_file, 'r') as fp:
        col_names = fp.readline().strip().split()[2:]
    # x axis is the first column, and y axis is the rest columns
    if fig_ax is None:
        fig, axs = plt.subplots(1, 2, figsize=(12, 4), constrained_layout=True)
    else:
        fig, axs = fig_ax
    axs[0].set_title(f'COLVAR @ TEMP {temp}K')
    axs[0].set_xlabel(r'$time / s$')
    for i, _col in enumerate(col_names[1:-1], start=1):
        axs[0].plot(colvar[:, 0], colvar[:, i])
    axs[0].legend(col_names[1:])
    axs[0].grid()

    # draw model_devi
    model_devi = np.loadtxt(model_devi_file, skiprows=1)  # the 4 colum is max_devi_f
    axs[1].set_title(f'Model Devi at @ TEMP {temp}K')
    axs[1].set_xlabel('step')
    axs[1].set_ylabel('max_devi_f')
    axs[1].plot(model_devi[:, 0], model_devi[:, 4])
    axs[1].grid()

    if save_to is None:
        fig.canvas.draw()
        fig.canvas.flush_events()
    else:
        fig.savefig(save_to, dpi=300, bbox_inches='tight')


class CmdEntries:
    def build_config(self):
        """
        Build config file for catalyst
        """
        return ConfigBuilder

# TODO: move this to dedicated ui module


def cli_main():
    fire.Fire(CmdEntries)


if __name__ == '__main__':
    fire.Fire(ConfigBuilder)


from ai2_kit.core.util import load_yaml_file
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import LiteralScalarString as LSS

from copy import deepcopy
import sys


class Yaml:
    def __init__(self) -> None:
        self.data = None

    def load(self, yaml_file: str):
        self.file = yaml_file
        self.data = load_yaml_file(yaml_file)
        return self

    def set_value(self, key: str, value):
        """edit a yaml file

        Args:
            yaml_file (str): path to the yaml file
            key (str): key to edit, support nested key, e.g. `a.b.c`
            value ([type]): new value
        """
        # TODO: handle missing key
        keys = key.split('.')
        d = self.data
        for k in keys[:-1]:
            d = d[k]
        d[keys[-1]] = value
        return self

    def dump(self, in_place = False, pretty=True):
        yaml = YAML()
        yaml.default_flow_style = False
        data = deepcopy(self.data)
        if pretty:
            _apply_lss(data)
        if in_place:
            with open(self.file, 'w') as fp:
                yaml.dump(data, fp)
        else: # to stdout
            yaml.dump(data, sys.stdout)


def _apply_lss(data: dict):
    """
    For each value in data, if it is a multiple line string, convert it to LSS.
    """
    for k, v in data.items():
        if isinstance(v, str) and '\n' in v:
            data[k] = LSS(v)
        elif isinstance(v, dict):
            _apply_lss(v)
        elif isinstance(v, list):
            for item in v:
                if isinstance(item, dict):
                    _apply_lss(item)